/* ----------------------------------------------------------
%   (C)1993,1994,1995 Institute for New Generation Computer Technology
%       (Read COPYRIGHT for detailed information.)
%   (C)1996, 1997, 1998, 1999 Japan Information Processing Development Center
%       (Read COPYRIGHT-JIPDEC for detailed information.)
----------------------------------------------------------- */

#include <math.h>
#include <stdio.h>
#include <klic/gdobject.h>  /* debug_printf, klic_fprintf */
#include <klic/gd_macro.h>

#include "atom.h"
#include "funct.h"

#ifdef DIST
#include "interpe.h"
#endif

#define CLASS_NAME  float
#define G_OBJ_TYPE  struct float_object
#define G_OBJ_SIZE  (G_SIZE_IN_Q(G_OBJ_TYPE))

#define ALIGN() \
do{ \
  if( sizeof(double) != sizeof(long) ){ \
    set_heapp((q*) (((unsigned long) heapp() + sizeof(double) - 1) \
		     & -sizeof(double) )); \
  } \
}while(0)

extern double atof();
extern q convert_c_string_to_klic_string();
extern unsigned char* generic_string_body();

#ifdef DIST
extern void push_decode_stack();
static void decode_float(combuf* inbuf);
#endif /* DIST */

GD_USE_CLASS(byte__string);

G_OBJ_TYPE {
  struct data_object_method_table *method_table;
  double value;
};

#ifdef DIST
union dl {
  double d;
  long l[2];
};
#endif /* DIST */

/* basic method definitions */

GDDEF_GUNIFY()
{
  G_STD_DECL;
  if( g_self->method_table != GD_OTHER->method_table ||
      g_self->value != GD_OTHER->value )
    GD_GFAIL;
  GD_GSUCCEED;
}

GDDEF_UNIFY()
{
  G_STD_DECL;

  if( g_self->method_table != GD_OTHER->method_table ||
      g_self->value != GD_OTHER->value )
    GD_UNIFY_FAIL;
}

G_DEF_GC()
{
  G_STD_DECL;
  G_OBJ_TYPE* newself;

  ALIGN();
  G_SET_NEWOBJ_IN_NEWGEN(newself);
  newself->value = g_self->value;
  G_RETURN_FROM_GC(newself);
}

/* Generic method */

GDDEF_METHOD(print_1)
{
  G_STD_DECL;
  double value = g_self->value;
  q str;
  char buf[100];
  sprintf(buf, "%1.16g", value);
  if( STRCHR(buf, '.') == NULL &&
      strcmp(buf, "Infinity") != 0 && strcmp(buf, "-Infinity") != 0 &&
      strcmp(buf, "NaN") != 0 ){
    char* exponent = STRCHR(buf, 'e');
    if( exponent != NULL ){
      char save[100];
      strcpy(save, exponent);
      strcpy(exponent, ".0");
      strcpy(exponent+2, save);
    }else{
      strcat(buf, ".0");
    }
  }
  str = convert_c_string_to_klic_string(buf);
  G_UNIFY_VALUE(g_argv[0], str);
}

GDDEF_METHOD(int_1)
{
  G_STD_DECL;
  double value = g_self->value;
  q result = makeint((long) value);
  G_UNIFY_VALUE(g_argv[0], result);
}

static double
unary_plus(x)
  double x;
{
  return x;
}

static double
unary_minus(x)
  double x;
{
  return -x;
}

/*  Generic Method Table */
GDDEF_GENERIC()
{
  G_STD_DECL;
  double self, result;
  unsigned long result_index;
  G_OBJ_TYPE* newobj;
  double (*func)();

  self = g_self->value;
  GD_SWITCH_ON_METHOD {

  GD_METHOD_CASE(print_1);
  GD_METHOD_CASE(int_1);
    break;

  default:
    ALIGN();
    GDSET_NEWOBJ(newobj);
    GD_SWITCH_ON_ARITY {
    case 1: {
      result_index = 0;
      GD_SWITCH_ON_METHOD {
	GD_METHOD_CASE_DIRECT(sin_1): func = sin; goto apply_1;
	GD_METHOD_CASE_DIRECT(cos_1): func = cos; goto apply_1;
	GD_METHOD_CASE_DIRECT(tan_1): func = tan; goto apply_1;
	GD_METHOD_CASE_DIRECT(asin_1): func = asin; goto apply_1;
	GD_METHOD_CASE_DIRECT(acos_1): func = acos; goto apply_1;
	GD_METHOD_CASE_DIRECT(atan_1): func = atan; goto apply_1;
	GD_METHOD_CASE_DIRECT(sinh_1): func = sinh; goto apply_1;
	GD_METHOD_CASE_DIRECT(cosh_1): func = cosh; goto apply_1;
	GD_METHOD_CASE_DIRECT(tanh_1): func = tanh; goto apply_1;
	GD_METHOD_CASE_DIRECT(exp_1): func = exp; goto apply_1;
	GD_METHOD_CASE_DIRECT(log_1): func = log; goto apply_1;
	GD_METHOD_CASE_DIRECT(sqrt_1): func = sqrt; goto apply_1;
	GD_METHOD_CASE_DIRECT(ceil_1): func = ceil; goto apply_1;
	GD_METHOD_CASE_DIRECT(floor_1): func = floor; goto apply_1;
	GD_METHOD_CASE_DIRECT(plus_1): func = unary_plus; goto apply_1;
	GD_METHOD_CASE_DIRECT(minus_1): func = unary_minus; goto apply_1;
	GD_METHOD_CASE_DEFAULT;
      }
    }
    case 2: {
      q another = g_argv[0];
      double another_value;
      GD_DEREF(another);
      result_index = 1;
      if( !isfunctor(another) ||
	  ((G_OBJ_TYPE*) functorp(another))->method_table !=
	  g_self->method_table ){
	debug_printf("### \"%k\" given to floating point method %F ###\n",
		     another, g_method_functor);
	fatal("Invalid argument in floating point object method.");
      }
      another_value = ((G_OBJ_TYPE*) functorp(another))->value;
      GD_SWITCH_ON_METHOD {
      GD_METHOD_CASE_DIRECT(add_2):
	result = self + another_value;
	goto apply_2;
      GD_METHOD_CASE_DIRECT(subtract_2):
	result = self - another_value;
	goto apply_2;
      GD_METHOD_CASE_DIRECT(multiply_2):
	result = self * another_value;
	goto apply_2;
      GD_METHOD_CASE_DIRECT(divide_2):
	result = self / another_value;
	goto apply_2;
      GD_METHOD_CASE_DIRECT(pow_2):
	result = pow(self, another_value);
	goto apply_2;
      GD_METHOD_CASE_DEFAULT;
      }
      break;
    }
    GD_METHOD_CASE_DEFAULT;
    }
  }
  return;

 apply_1:
  result = func(self);
 apply_2:
  newobj->value = result;
  do_shallow_unify(g_argv[result_index], makefunctor(newobj));
}

/* guard generic methods */

#define COMPARE_METHOD(comparison)				\
{								\
  G_STD_DECL;							\
  q otherq = g_argv[0];					\
  G_OBJ_TYPE* other;						\
  double self, theother;					\
								\
  if( !G_ISGOBJ(otherq) ) GD_GFAIL;				\
  other = (G_OBJ_TYPE*) functorp(otherq);			\
  if( other->method_table != g_self->method_table ) GD_GFAIL;	\
  self = g_self->value;					\
  theother = other->value;					\
  if( comparison ) GD_GSUCCEED;					\
  GD_GFAIL;							\
}

GDDEF_GMETHOD(less__than_1)
{ COMPARE_METHOD(self < theother); }

GDDEF_GMETHOD(not__greater__than_1)
{ COMPARE_METHOD(self <= theother); }

GDDEF_GMETHOD(not__less__than_1)
{ COMPARE_METHOD(self >= theother); }

GDDEF_GMETHOD(greater__than_1)
{ COMPARE_METHOD(self > theother); }

GDDEF_GMETHOD(equal_1)
{ COMPARE_METHOD(self == theother); }

GDDEF_GMETHOD(not__equal_1)
{ COMPARE_METHOD(self != theother); }

GDDEF_GMETHOD(float_0)
{
  G_STD_DECL;
  GD_GSUCCEED;
}

GDDEF_GGENERIC()
{
  G_STD_DECL;
  GD_SWITCH_ON_METHOD {
    GD_GMETHOD_CASE(less__than_1);
    GD_GMETHOD_CASE(not__greater__than_1);
    GD_GMETHOD_CASE(not__less__than_1);
    GD_GMETHOD_CASE(greater__than_1);
    GD_GMETHOD_CASE(equal_1);
    GD_GMETHOD_CASE(not__equal_1);
    GD_GMETHOD_CASE(float_0);
    GD_GMETHOD_CASE_DEFAULT;
  }
}

G_DEF_PRINT()
{
  G_STD_DECL;
  char b[100];
  sprintf(b, "%f", g_self->value);
  klic_fprintf(g_fp, "%s", b);
  G_RETURN_FROM_PRINT();
}

GDDEF_COMPARE()
{
  G_STD_DECL;
  if( g_self->value > GD_OTHER->value ){
    return makeint(1);
  }else if( g_self->value < GD_OTHER->value ){
    return makeint(-1);
  }else{
    return makeint(0);
  }
}

GDDEF_HASH()
{
  G_STD_DECL;
  return makeint((long) g_self->value);
}

#ifdef DIST
G_DEF_ENCODE()
{
  G_STD_DECL;

  PUT_BUFFER(buffer, decode_float);
  if( sizeof(double) == sizeof(long) ){
    PUT_BUFFER(buffer, (long) g_self->value);
  }else{
    union dl dl;
    dl.d = g_self->value;
    PUT_BUFFER(buffer, dl.l[0]);
    PUT_BUFFER(buffer, dl.l[1]);
  }
  return GENERIC_SUCCEEDED;
}
#endif /* DIST */

#define GDUSE_MY_GUNIFY
#define GDUSE_MY_UNIFY
#define GDUSE_MY_PRINT
#define GDUSE_MY_GC
#ifdef DIST
#define GDUSE_MY_ENCODE
#endif
#define GDUSE_MY_GENERIC
#define GDUSE_MY_GGENERIC
#define GDUSE_MY_COMPARE
#define GDUSE_MY_HASH

/* define the method table structure of the vector */

#include <klic/gd_methtab.h>

/*  new_float function */
G_DEF_NEW()
{
  G_STD_DECL_FOR_NEW();
  q init = g_argv[0];
  G_OBJ_TYPE* newobj;

  G_DEREF_FOR_NEW(init);
  ALIGN();
  G_SET_NEWOBJ_FOR_NEW(newobj, G_SIZE_IN_Q(G_OBJ_TYPE));
  if( isint(init) ){
    newobj->value = (double) intval(init);
  }else if( G_ISGOBJ(init) &&
	    (struct data_object_method_table*) functor_of(init)
	    == &byte__string_g_method_table ){
    newobj->value = atof((char*) generic_string_body(functorp(init)));
  }else{
    debug_printf("### %k ###\n", init);
    fatal("Invalid initial value for floating point objects");
  }
  GD_RETURN_FROM_NEW(newobj);
}

#ifdef DIST
static void
decode_float(inbuf)
  combuf* inbuf;
{
  G_STD_DECL;
  G_OBJ_TYPE* newobj;

  ALIGN();
  newobj = (G_OBJ_TYPE*) klic_alloc(G_SIZE_IN_Q(G_OBJ_TYPE));

  newobj->method_table = &G_method_table;
  if( sizeof(long) == sizeof(double) ){
    newobj->value = (double) GET_BUFFER(inbuf);
  }else{
    union dl dl;
    dl.l[0] = (long) GET_BUFFER(inbuf);
    dl.l[1] = (long) GET_BUFFER(inbuf);
    newobj->value = dl.d;
  }
  push_decode_stack((q) makefunctor(newobj));
}
#endif /* DIST */
